/**
 * \file Comparator.hxx Defines virtual Comparator interfaces and an implementation, PropertySetComparator
 * \author Li Li
 * \date 4/30/08
 */
#ifndef Comparator_HXX_
#define Comparator_HXX_

#include <string>
#include <map>
#include "Metadata.hxx"

#define LT		1301
#define LTEQ	1302	
#define EQ		1303	
#define NEQ		1304	
#define GT		1305
#define GTEQ	1306
#define BT		1307

#define LEFT	110
#define RIGHT	111

using namespace std;

namespace cqosdb 
{ 
	/**
      *  Virtual Comparator interfaces that specifies comparison operations between parameter sets 
      */
	class Comparator {
		/* Comparison operations between parameter sets */
		public:
		virtual ~Comparator(){}
		
		/**
			Return class name
		*/
		virtual string getClassName()=0;
		
		/**
			Set left hand side parameter set to be compared
		*/
		virtual	void setLHS(ParameterSet& lefthand)=0;
		
		/**
			Set right hand side parameter set to be compared
		*/
		virtual	void setRHS(ParameterSet& righthand)=0;
		
		/**
			Grab left hand side parameter set to be compared
		*/
		virtual ParameterSet& getLHS()=0;
		
		/**
			Grab right hand side parameter set to be compared
		*/
		virtual ParameterSet& getRHS()=0;
			
		/**
			Grab a parameter at the left hand side of comparison
			\param paraName Name string of the parameter
		*/
		virtual Parameter* getLHSParameterAt(string paraName)=0;
		
		/**
			Grab a parameter at the right hand side of comparison
			\param paraName Name string of the parameter
		*/
		virtual	Parameter* getRHSParameterAt(string paraName)=0;
			
		/**
			Get count of parameters to be compared
		*/
		virtual	int getDimension()=0;
		
		/**
			Set up comparison tolerance of a parameter
			\param name Parameter name
			\param epsilon Comparison tolerance value
		*/
		virtual void setToleranceAt(string name, double epsilon)=0;
		
		/**
			Get comparison tolerance value of a parameter.
			\param name Parameter name
		*/
		virtual double getToleranceAt(string name)=0;
			
		/**
			Set comparison relation at a parameter
			\param name Parameter name
			\param aRelation Comparison relation
		*/
		virtual void setRelationAt(string name, int aRelation)=0;
		
		/**
			Grab comparison relation at a parameter
			\param name Parameter name
		*/
		virtual int getRelationAt(string name)=0;
		
		/**
			After setting up left, right hand side, tolerances, and comparison relation, do comparison
		*/
		virtual	bool doCompare()=0;		
	};


	/**
	  * Implement Comparator interfaces for comparing matrix property sets.
	  */
	class PropertySetComparator:
		public virtual Comparator
	{
	
	public:
		/**
		Constructor: initialize with left hand side, right hand side, comparison tolerances and relations
		\param LHS left hand side of comparison
		\param RHS right hand side of comparison
		\param tMap comparison tolerances
		\param rMap comparison relations
		*/
		PropertySetComparator(PropertySet& LHS, PropertySet& RHS, map<string, double>& tMap, map<string, int>& rMap);
		
		/**
		Constructor: initialize with left hand side, right hand side, comparison relations
		\param LHS left hand side of comparison
		\param RHS right hand side of comparison
		\param rMap comparison relations
		*/
		PropertySetComparator(PropertySet& LHS, PropertySet& RHS, map<string, int>& rMap);

		/**
		Constructor: initialize with left hand side andright hand side
		\param LHS left hand side of comparison
		\param RHS right hand side of comparison
		*/
		
		PropertySetComparator(PropertySet& LHS, PropertySet& RHS);
		PropertySetComparator(){};
		~PropertySetComparator();
		string getClassName();
		void clear();
		
		void setLHS(ParameterSet& lefthand); 
		void setRHS(ParameterSet& righthand); 
		
		ParameterSet& getLHS();
		ParameterSet& getRHS();
		
		Parameter* getLHSParameterAt(string paraName);
		Parameter* getRHSParameterAt(string paraName);
					
		int getDimension();
		
		void setToleranceAt(string name, double epsilon);
		double getToleranceAt(string name);
		
		void setRelationAt(string name, int aRelation);
		int getRelationAt(string name);
	
		/**
			\return true if (LHS[i] - RHS[i]) rMap[i] tolerance[i] is true for every property i (i>=0 && i< parameter count); false otherwise.					
		*/
		bool doCompare();			
			
		//PropertySetComparator &operator=(const PropertySetComparator& another);

	private:
		PropertySet lhs, rhs;
		map<string, double> toleranceMap;
		map<string, int> relationMap;
			
	};


}  // namespace cqosdb
#endif


