
#include "MetaDB.hxx"
#include <stdlib.h>
#include <iostream>
#include <pgtypes_numeric.h>
#include <pgtypes_error.h> 

#define exit_nicely(conn) \
{\
    PQfinish(conn);\
    exit(1);\
}

#define CHKERR(ierr) if(ierr<0) return ierr;

namespace cqosdb {

	MetaDB::MetaDB() 
	{
		connected = -1;
		paramCount = 0;
		conninfo = NULL;
		conn = NULL;
		for (int index=0; index<MAXPARAMCOUNT; index++)
			paramValues[index] = NULL;
	}
	
	MetaDB::~MetaDB() 
	{
		if(!(isClosed())){
		disconnect();}
	}

	bool MetaDB::connect() 
	{
		if (connected != -1) {
			return true;
		}
		
		if (conninfo == NULL){
			printf("Provide user access info first by calling setConnectionInfo()");
			return false;
		}
		
		conn = PQconnectdb(conninfo);
		
		if (PQstatus(conn) != CONNECTION_OK)
		{
			fprintf(stderr, "Connection to database failed: %s",
                PQerrorMessage(conn));
			exit_nicely(conn);
		}
		
		connected = 1; 		
		return true;
        
	}

	void MetaDB::setConnectionInfo(string info){
		conninfo = info.c_str();
	}
	
	string MetaDB::getConnectionInfo(){
		return string(conninfo);
	}
	
	
	string MetaDB::getConnectionInfoFromDefaultPerfDMFConfig(){
		string fileName(getenv("HOME"));
		fileName += "/.ParaProf/perfdmf.cfg";
		return getConnectionInfoFromPerfDMFConfig(fileName);
	}
	
	string MetaDB::getConnectionInfoFromPerfDMFConfig(string configFileName){
		string tempStr = "";
		ConfigFile config(configFileName);
		tempStr += "host=";
		tempStr += config.read<string>("db_hostname");
		tempStr += " port=";
		tempStr += config.read<string>("db_portnum");
		tempStr += " user="; 
		tempStr += config.read<string>("db_username");
		tempStr += " password=";
		tempStr += config.read<string>("db_password");
		tempStr += " dbname=";
		tempStr += config.read<string>("db_dbname");
		
		conninfo = tempStr.c_str();
	
		return tempStr;
	}
	
	void MetaDB::disconnect() 
	{	
	    if (isClosed()) 
	    	return;
	    else 
	    {
	    	connected = -1;
			PQfinish(conn);
	    }
	}
	
	bool MetaDB::isClosed(){
		if (connected == -1)
			return true;
		return false;
	}

	void MetaDB::bindCommand()
	{
		if(stringarg.size()>0) {
			map<int,string>::iterator it;
			for(it = stringarg.begin(); it!=stringarg.end(); ++it) {
				paramValues[paramCount] = it->second.c_str();
				paramCount++;
			}
		}	
	}

	
	int MetaDB::bindParameters(int param,int argument)
	{
		int ierr = bindParameters(param,Utility::T_as_string<int>(argument));
		return ierr;
	}

//---------------------------------------------------------------------------
    int MetaDB::bindParameters(int param,long argument)
    {
	int ierr = bindParameters(param, Utility::T_as_string<long>(argument));
	return ierr;
    }
	
//---------------------------------------------------------------------------
    int MetaDB::bindParameters(int param,double argument)
    {
	int ierr = bindParameters(param, Utility::T_as_string<double>(argument));
	return ierr;
    }
	
//---------------------------------------------------------------------------
    int MetaDB::bindParameters(int param,string argument)
    {
	if(stringarg.find(param)!=stringarg.end())
	{ 
		fprintf(stderr,"Parameter %d already set!\n",param);
		fflush(stderr); return -1;
	}
	
	stringarg[param]=argument;
	return 0;
    }
	
	void MetaDB::resetQuery(){
		paramCount = 0;
		stringarg.clear();	
	}


	int MetaDB::executeQuery_bindParam(string& commd, Outcome* res)  
    {
	   if ((connected == -1) || (PQstatus(conn) != CONNECTION_OK)) {
			string mssg(" Database is closed for ");
			mssg = mssg + commd;
			fprintf(stderr, mssg.c_str());
			fflush(stderr); return -1;
	    }

		PGresult *results;
		
		bindCommand();
		
		results = PQexecParams(conn,
						commd.c_str(),
						paramCount,
						NULL,    /* let the backend deduce param type */
						paramValues,
						NULL,    /* don't need param lengths since text */
						NULL,    /* default to all text params */
						0);      /* ask for text results */ 
										
		res->setElements(results);
		PQclear(results);
		
		return 0;
    }

	int MetaDB::executeQuery(string& commd, Outcome *res)  
    {
	   if ((connected == -1) || (PQstatus(conn) != CONNECTION_OK)) {
			string mssg(" Database is closed for ");
			mssg = mssg + commd;
			fprintf(stderr, mssg.c_str());
			fflush(stderr); return -1;
	    }

		PGresult *results;
		
		results = PQexec(conn, commd.c_str());
		res->setElements(results);
		PQclear(results);
		
		return 0;
    }

	int MetaDB::executeInsert(string& commd)  
    {
	   if ((connected == -1) || (PQstatus(conn) != CONNECTION_OK)) {
			string mssg(" Database is closed for ");
			mssg = mssg + commd;
			fprintf(stderr, mssg.c_str());
			fflush(stderr); return -1;
	    }

		PGresult *results;
		
		bindCommand();
		
		results = PQexecParams(conn,
						commd.c_str(),
						paramCount,
						NULL,    /* let the backend deduce param type */
						paramValues,
						NULL,    /* don't need param lengths since text */
						NULL,    /* default to all text params */
						0);      /* ask for text results */ 
										
		
		if (PQresultStatus(results) != PGRES_COMMAND_OK)
		{
			fprintf(stderr, "command failed: %s\n", PQerrorMessage(conn));
			PQclear(results);
			exit_nicely(conn);
		}

		PQclear(results);
		
		return 0;
    }

	
	int MetaDB::getMatchingTrialsBetween(ParameterSet& lower, ParameterSet& upper, Outcome* trialIDs){ 
	
		int paramsetSize=lower.getNumOfParameters(), counter=0;
		string select=" ", conditions=" ", query;
		double leftV, rightV;
		
		if ((lower.getClassName().compare("PropertySet") ==0) && (lower.getClassName().compare(upper.getClassName()) ==0)){ 
			select += "select DISTINCT t" + Utility::T_as_string<int>(counter) + ".trial,  ";
			
			for (int index=0; index<paramsetSize; index++){
				select += "t" + Utility::T_as_string<int>(counter) + ".parameter_type, t" + Utility::T_as_string<int>(counter) + ".parameter_value";
				
				if (index != paramsetSize-1) select += ", ";
				
				counter++;
			}
			
			select += " from ";
			
			counter = 0;
			for (int index=0; index<paramsetSize; index++){	
				select += lower.getTBName() + " t" + Utility::T_as_string<int>(counter);
				if (counter != paramsetSize-1)
					select += ",";
					
				counter++;
			}
			
			counter = 0;
			
			for (int index=0; index<paramsetSize; index++){
				
				if (counter != 0)
					conditions += " AND ";
					
				conditions += "t" + Utility::T_as_string<int>(counter) + ".parameter_name=\'" + (lower.getParameterAt(index))->getName() + "\' ";
							  
				if (index < paramsetSize-1){
					conditions += " AND t" + Utility::T_as_string<int>(counter) +".trial=t" + Utility::T_as_string<int>(counter+1) + ".trial"
								  + " AND t" + Utility::T_as_string<int>(counter) +".interval_event=t" + Utility::T_as_string<int>(counter+1) + ".interval_event";
				}
				
				counter++;
			}
			
			query = select + " where " + conditions;
			//cout << query << endl;
			
			ResultSet Res;
			executeQuery(query, &Res);
			
			int len = Res.getNumberOfRows();
						
			string type, value;
			bool flag = true;
			
			if (len>0) {
			
				// init trialIDs
				ResultSet* tempRes = dynamic_cast<ResultSet*>(trialIDs);
				tempRes->setColumnName(Res.getColumnName());
				tempRes->setParamTypes(Res.getParameterTypes(), Res.getNumberOfColumns() );
			
				for(int i=0;i<len;++i)
				{
					for (int index=0; index<paramsetSize; index++){
						Res.getElement(i,1+index*2,&type);
						Res.getElement(i,1+index*2+1,&value);
						
						Utility::stringToLower((char *)type.c_str(),type.size());
				
						if (type.compare("int")==0){
							int iValue = Utility::string_as_T<int>(value);
							(lower.getParameterAt(index))->getValue(&leftV);
							(upper.getParameterAt(index))->getValue(&rightV);
							if ((iValue < leftV) || (iValue > rightV)){
								flag = false;
								break;
							}
						}
						else if (type.compare("long")==0){
							long lValue = Utility::string_as_T<long>(value);
							(lower.getParameterAt(index))->getValue(&leftV);
							(upper.getParameterAt(index))->getValue(&rightV);
							if ((lValue < leftV) || (lValue > rightV)){
								flag = false;
								break;
							}
						}
						else if (type.compare("double")==0){
							double dValue = Utility::string_as_T<double>(value);
							(lower.getParameterAt(index))->getValue(&leftV);
							(upper.getParameterAt(index))->getValue(&rightV);
							if ((dValue < leftV) || (dValue > rightV)){
								flag = false;
								break;
							}
						}
						else {
							fprintf(stderr,"Not support the data type yet!\n");fflush(stderr);
							return -1;
						}	
						
					}
					
					if (flag){ // add it to trialIDs
						int trialid, result;
						bool tempFlag = true;
						Res.getElement(i,"trial",&result);
						for(int resIndex=0;resIndex<tempRes->getNumberOfRows();++resIndex)
						{
							
							tempRes->getElement(resIndex, "trial", &trialid); 
							if (trialid==result){ // the trial id's already added
								tempFlag = false;
								break;
							}
						}
						
						if (tempFlag) tempRes->addARow(Res.getARow(i));
					}
				}
				
				return 1;
			}			
		}
		
		cout << "Unmatched parameter set type to be compared.\n" ;
		return -1;
	}
		
	// query DB with trials whose parameter set is within [lower-epsilon, lower+epsilon]
	int MetaDB::getMatchingTrials(ParameterSet& lower, vector<double> epsilons, Outcome* trialIDs){
	
		int paramsetSize=lower.getNumOfParameters(), counter=0;
		string select=" ", conditions=" ", query;
		double leftV;
		
		if (lower.getClassName().compare("PropertySet") ==0){
			select += "select DISTINCT t" + Utility::T_as_string<int>(counter) + ".trial, ";
			
			for (int index=0; index<paramsetSize; index++){
			
				select += "t" + Utility::T_as_string<int>(counter) + ".parameter_type, t" + Utility::T_as_string<int>(counter) + ".parameter_value";
				
				if (index != paramsetSize-1) select += ", ";
				
				counter++;
			}
			select += " from ";
			
			counter = 0;
			for (int index=0; index<paramsetSize; index++){	
				select += lower.getTBName() + " t" + Utility::T_as_string<int>(counter);
				if (counter != paramsetSize-1)
					select += ",";
					
				counter++;
			}

			counter =0;
			
			for (int index=0; index <paramsetSize; index++){
				
				if (counter != 0)
					conditions += " AND ";
					
				conditions += "t" + Utility::T_as_string<int>(counter) + ".parameter_name=\'" + (lower.getParameterAt(index))->getName() + "\' ";
							  				
				if (index < paramsetSize-1){
					conditions += " AND t" + Utility::T_as_string<int>(counter) +".trial=t" + Utility::T_as_string<int>(counter+1) + ".trial"
								  + " AND t" + Utility::T_as_string<int>(counter) +".interval_event=t" + Utility::T_as_string<int>(counter+1) + ".interval_event";
				}
				
				counter++;
			}
			
			query = select + " where " + conditions;
			
			//cout << query << "\n";
			
			ResultSet Res;
			executeQuery(query, &Res);

			int len = Res.getNumberOfRows();
						
			string type, value;
			bool flag;
			
			if (len>0) {

				// init trialIDs
				ResultSet* tempRes = dynamic_cast<ResultSet*>(trialIDs);
				tempRes->setColumnName(Res.getColumnName());
				tempRes->setParamTypes(Res.getParameterTypes(), Res.getNumberOfColumns() );
			
				for(int i=0;i<len;++i)
				{
					flag = true;
					
					for (int index=0; index<paramsetSize; index++){
						(lower.getParameterAt(index))->getValue(&leftV);

						Res.getElement(i,1+index*2,&type);
						Res.getElement(i,1+index*2+1,&value);
						
						Utility::stringToLower((char *)type.c_str(),type.size());
				
						if (type.compare("int")==0){
							int iValue = Utility::string_as_T<int>(value);
							if ((iValue < leftV-epsilons[index]) || (iValue > leftV+epsilons[index])){
								flag = false;
								break;
							}
						}
						else if (type.compare("long")==0){
							long lValue = Utility::string_as_T<long>(value);
							
							if ((lValue < leftV-epsilons[index]) || (lValue > leftV+epsilons[index])){
								flag = false;
								break;
							}
						}
						else if (type.compare("double")==0){
							double dValue = Utility::string_as_T<double>(value);
							
							if ((dValue < leftV-epsilons[index]) || (dValue > leftV+epsilons[index])){
								flag = false;
								break;
							}
						}
						else {
							fprintf(stderr,"Not support the data type yet!\n");fflush(stderr);
							return -1;
						}	
						
					}

					if (flag){ // add it to trialIDs
						int trialid, result;
						bool tempFlag = true;
						Res.getElement(i,"trial",&result);

						for(int resIndex=0;resIndex<tempRes->getNumberOfRows();++resIndex)
						{
							
							tempRes->getElement(resIndex, "trial", &trialid); 
							if (trialid==result){ // the trial id's already added
								tempFlag = false;
								break;
							}
						}
						
						if (tempFlag) tempRes->addARow(Res.getARow(i));
					}
				}
				
				return 1;
			}			

		}	
		
		cout << "Unknown parameter set type.\n" ;
		return -1;
	}

			
	void MetaDB::getParameter(int trialID, string interEvt, Parameter* aParam){ 
	// get parameter value from DB, assume parameter name and table name have already set in aParam
	
			string query;
			
			if (aParam->getClassName().compare("DefaultProperty") ==0){
				query = "select  "+aParam->getTBName()+ ".parameter_value, " + aParam->getTBName()+ ".parameter_type " + "from " + aParam->getTBName() + " where "
						+ aParam->getTBName() + ".trial= " + Utility::T_as_string<int>(trialID) + " AND "
						+ aParam->getTBName() + ".interval_event= \'" + interEvt + "\' AND "
						+ aParam->getDBConditions();
				//cout<<query;
			}
			else {
				cout << "Unknown parameter type.\n" ;
				return;
			}
			
			ResultSet res;
			executeQuery(query, &res);
			if (res.getNumberOfRows() > 1){
				printf("There are multiple values with trialID: %d, interEvt: %s, property name: %s. We assign the parameter the first returned value.\n", trialID, interEvt.c_str(), aParam->getName().c_str());
			}
			
			
			string type, value;
			res.getElement(0,0,&value);
			res.getElement(0,1,&type);
			
			Utility::stringToLower((char *)type.c_str(),type.size());
			
			if (type.compare("string")==0)
				aParam->setValue(value);
			else if (type.compare("boolean")==0){
				bool bValue = Utility::string_as_T<bool>(value);
				aParam->setValue(bValue);
			}
			else if (type.compare("int")==0){
				int iValue = Utility::string_as_T<int>(value);
				aParam->setValue(iValue);
			}
			else if (type.compare("long")==0){
				long lValue = Utility::string_as_T<long>(value);
				aParam->setValue(lValue);
			}
			else if (type.compare("double")==0){
				double dValue = Utility::string_as_T<double>(value);
				aParam->setValue(dValue);
			}
			else {
				fprintf(stderr,"Not support the data type yet!\n");fflush(stderr);
				return;
			}
			
	}
	
	void MetaDB::getParameterSet(int trialID, string interEvt, ParameterSet* aParamSet){	// get parameter set values	
	
		string select=" ", conditions=" ", query;
		int paramsetSize = aParamSet->getNumOfParameters();
			  
		if (aParamSet->getClassName().compare("PropertySet") ==0){

			select += "select ";
			
			for (int index=0; index<paramsetSize; index++){
			
				select += "t" + Utility::T_as_string<int>(index) + ".parameter_name as p"+ Utility::T_as_string<int>(index) +"_name, " 
						+ "t" + Utility::T_as_string<int>(index) + ".parameter_type as p"+ Utility::T_as_string<int>(index) +"_type, "
						+ "t" + Utility::T_as_string<int>(index) + ".parameter_value as p" + Utility::T_as_string<int>(index) + "_value";
							
				if (index != paramsetSize-1)
					select += ",";
				
			}
			
			select += " from ";
			
			for (int index=0; index<paramsetSize; index++){
				select += aParamSet->getTBName() + " t" + Utility::T_as_string<int>(index);
				if (index != paramsetSize-1)
					select += ",";
			}
			
			
			for (int index=0; index<paramsetSize; index++){
				if (index != 0)
					conditions += " AND ";
				
				conditions += "t" + Utility::T_as_string<int>(index) + ".trial=" + Utility::T_as_string<int>(trialID);
				conditions += " AND ";
				conditions += "t" + Utility::T_as_string<int>(index) + ".interval_event= \'" + interEvt + "\' ";
				conditions += " AND ";
				conditions += "t" + Utility::T_as_string<int>(index) + "." + "parameter_name = \'" + (aParamSet->getParameterAt(index))->getName() + "\'";
			}
			
			query = select + " where " + conditions;
			
			//cout << query <<"\n";
		}
		else {
				cout << "Unknown parameter set type.\n" ;
				return;
		}
			
		ResultSet res;
		executeQuery(query, &res);
				
		if (res.getNumberOfRows() > 1){
				printf("There are multiple values with trialID: %d, interEvt: %s, and the parameter set. We assign the parameter set the first returned value.\n", trialID, interEvt.c_str());
		}

		string type, value;
		
		for (int index =0; index< paramsetSize; index++){
			res.getElement(0,index*3+1,&type);
			res.getElement(0,index*3+2,&value);
			
			Utility::stringToLower((char *)type.c_str(),type.size());
			
			if (type.compare("string")==0)
				(aParamSet->getParameterAt(index))->setValue(value);
			else if (type.compare("boolean")==0){
				bool bValue = Utility::string_as_T<bool>(value);
				(aParamSet->getParameterAt(index))->setValue(bValue);
			}
			else if (type.compare("int")==0){
				int iValue = Utility::string_as_T<int>(value);
				(aParamSet->getParameterAt(index))->setValue(iValue);
			}
			else if (type.compare("long")==0){
				long lValue = Utility::string_as_T<long>(value);
				(aParamSet->getParameterAt(index))->setValue(lValue);
			}
			else if (type.compare("double")==0){
				double dValue = Utility::string_as_T<double>(value);
				(aParamSet->getParameterAt(index))->setValue(dValue);
			}
			else {
				fprintf(stderr,"Not support the data type yet!\n");fflush(stderr);
				return;
			}

		}
		
	}
	
	void MetaDB::storeParameter(int trialID, string interEvt, Parameter& aParam){
		string tbName = aParam.getTBName(), query = "insert into ";
		string value;
		double dvalue;
		bool bvalue;
		
		if (aParam.getClassName().compare("DefaultProperty") ==0){
			DefaultProperty& castP = dynamic_cast<DefaultProperty&>(aParam); 
			query += tbName + " values (" + Utility::T_as_string<int>(trialID) + ", "
					+ "\'" + interEvt + "\'" + ", ";
			if (!castP.getCatName().empty())
				query += "\'"+ castP.getCatName() + "\', ";
			else
				query += "NULL, ";
				
			query += "\'"+castP.getName() + "\', ";
			
			if (!castP.getParamType().empty())
				query += "\'"+ castP.getParamType() + "\', ";
			else
				query += "NULL, ";
				
			if (castP.getParamType().compare("string")==0){
				castP.getValue(&value);
				query += "\'" + value + "\')";
			} 
			else if (castP.getParamType().compare("bool")==0){
				castP.getValue(&bvalue);
				query += "\'" + Utility::T_as_string<bool>(bvalue) + "\')";
			}
			else if ((castP.getParamType().compare("int")==0) || (castP.getParamType().compare("long")==0) || (castP.getParamType().compare("double")==0)){
				castP.getValue(&dvalue);
				query += Utility::T_as_string<double>(dvalue) + ")";
			}
			else {// the parameter has no value
				query += "NULL)";
			}
		}
		else {
				cout << "Unknown parameter type.\n" ;
				return;
		}
		
		cout << query << endl;
		executeInsert(query);
	}
	
	void MetaDB::storeParameterSet(int trialID, string interEvt, ParameterSet& aParamSet){
	
		if (aParamSet.getClassName().compare("PropertySet") ==0){
			int size = aParamSet.getNumOfParameters();
			
			for (int index=0; index<size; index++){
				storeParameter(trialID, interEvt, *aParamSet.getParameterAt(index));
			}
		}
		else {
				cout << "Unknown parameter set type.\n" ;
				return;
		}

	}
	
	/************* End of MetaDB  *****************/
	
	ResultSet::ResultSet()
	{
		reserved = -1; 
		rows = 0;
		parameter_type = NULL;
	}

	ResultSet::ResultSet(const ResultSet& other){

		QueryData = other.QueryData;
		parameter_type = (int *)malloc(sizeof(int)*other.columns);	
       
	   if(parameter_type == NULL)
       {
		fprintf(stderr,"MEMORY ALLOCATION FAILED!\n");fflush(stderr);
		return;
       } 
	
		for (int i=0; i< other.columns; i++)
			parameter_type[i] = other.parameter_type[i];
			
		reserved = other.reserved;
		column_name = other.column_name;
		rows = other.rows;
		columns = other.columns;
	}
	
	ResultSet::~ResultSet()
	{
		clear(); 
	}

	void ResultSet::clear()
	{
		for(int iv=0; iv < QueryData.size();++iv)
			QueryData[iv].clear();
		column_name.clear();
		QueryData.clear();
		if(reserved > 0 )
		{
			free(parameter_type);reserved = -1;
		}
		rows = 0;
	}

	int ResultSet::getNumberOfRows()
	{
		int retval = QueryData.size();
	
		if(retval != rows)
		{
			fprintf(stderr,"Internal FATAL ERROR\n");
			fflush(stderr); return -1;
		}	
		return rows;
	}



	int ResultSet::getNumberOfColumns()
	{
		return columns;
	}


	void  ResultSet::setColumnName(map<string,int> colName){		
		column_name = colName;
	}
	
		void ResultSet::setParamTypes( int* paramTypes, int colCount){
			parameter_type = (int *)malloc(sizeof(int)*colCount);
			
			columns = colCount;
			reserved = colCount;
			
			if(parameter_type == NULL)
		   {
			fprintf(stderr,"MEMORY ALLOCATION FAILED!\n");fflush(stderr);
			return;
		   } 
		
			for (int i=0; i< colCount; i++)
				parameter_type[i] = paramTypes[i];
		}
		
		
	int ResultSet::getElement(int row,int column,double *res)
	{

		if(rows <= row)
		{
			fprintf(stderr, "ERROR: Row number %d bigger than the number\
					of existing rows %d\n",row,rows);
			fflush(stderr); return -1;
		}
		
		if(reserved == -1 || column >= reserved)
		{
			fprintf(stderr, "ERROR: Column number %d bigger than the number\
				of existing types %d\n",column,reserved);
			fflush(stderr); return -1;
		}

		int ierr = QueryData[row].getElement(column, res,parameter_type[column]);
		return ierr;	

	}

	int ResultSet::getElement(int row,int column,long *res)
	{
		if(rows <= row)
		{
			fprintf(stderr, "ERROR: Row number %d bigger than the number\
		of existing rows %d\n",row,rows);
			fflush(stderr); return -1;
		}
		
		if(reserved == -1 || column >= reserved)
		{
			fprintf(stderr, "ERROR: Column number %d bigger than the number\
					of existing types %d\n",column,reserved);
			fflush(stderr); return -1;
		}
	
		int ierr = QueryData[row].getElement(column, res,parameter_type[column]);
		return ierr;	
	}
	

	int ResultSet::getElement(int row,int column,string *res)
	{
		if(rows <= row)
		{
			fprintf(stderr, "ERROR: Row number %d bigger than the number\
					of existing rows %d\n",row,rows);
			fflush(stderr); return -1;
		}
		int ierr = QueryData[row].getElement(column, res);
		return ierr;	
	}

	int ResultSet::getElement(int row,int column,void *res)
	{
		int ierr;
		if(parameter_type[column] == META_DOUBLE)
			ierr = getElement(row,column,(double *)res);
		else
		if(parameter_type[column] == META_LONG)
			ierr = getElement(row,column,(long *)res);
		else
		if(parameter_type[column] == META_STRING)
			ierr = getElement(row,column,(string *)res);
		else
		{
			fprintf(stderr,"ERROR:Unknown parameter type %d!\n",
			parameter_type[column]);fflush(stderr); ierr = -1;
		}
		return ierr;
	}

	int ResultSet::getElement(int row,string name,double *res)
	{

		Utility::stringToLower((char *)name.c_str(),name.size());
		if(column_name.find(name) == column_name.end())
		{
			fprintf(stderr, "ERROR: No column named %s\n", name.c_str());
			fflush(stderr); return -1;
		}
		int ierr = getElement(row, column_name[name],res);
		return ierr;

	}


	int ResultSet::getElement(int row,string name,long *res)
	{
		Utility::stringToLower((char *)name.c_str(),name.size());
		if(column_name.find(name) == column_name.end())
		{
			fprintf(stderr, "ERROR: No column named %s\n", name.c_str());
			fflush(stderr); return -1;
		}
		int ierr = getElement(row, column_name[name],res);
		return ierr;	
	}


	int ResultSet::getElement(int row,string name,string *res)
	{
		Utility::stringToLower((char *)name.c_str(),name.size());
		if(column_name.find(name) == column_name.end())
		{
			fprintf(stderr, "ERROR: No column named %s\n", name.c_str());
			fflush(stderr); return -1;
		}
		int ierr = getElement(row, column_name[name],res);
		return ierr;	
	}

int ResultSet::getElement(int row,string name,void *res)
{

	Utility::stringToLower((char *)name.c_str(),name.size());
	if(column_name.find(name) == column_name.end())
	{
		fprintf(stderr, "ERROR: No column named %s\n", 
			name.c_str());
		fflush(stderr); return -1;
	}
	int ierr = getElement(row, column_name[name],res);
	return ierr;
}

int ResultSet::setElements(PGresult *res)
{
       int col_count = PQnfields(res);
       int index = 0, ierr;
       
       clear();

       parameter_type = (int *)malloc(sizeof(int)*col_count);	
       if(parameter_type == NULL)
       {
		fprintf(stderr,"MEMORY ALLOCATION FAILED!\n");fflush(stderr);
		return -1;
       } 
	   	
       reserved = col_count; 
	
		ierr = setParameterType(res);
		CHKERR(ierr);

		ierr = setParameterName(res);
		CHKERR(ierr);
		
		for (index=0; index<PQntuples(res); index++){
			
			ColumnData Col; 
			for(int i=0;i<col_count;++i)
			{
				//ierr = Col.setElement(&(cmd->Field(i+1)),
				//i,parameter_type[i]); CHKERR(ierr);
				ierr = Col.setElement(res, index, i, parameter_type[i]); 
				CHKERR(ierr);
			}
			QueryData.push_back(Col);
		}
	   
	    rows = QueryData.size();
		columns = col_count;
		return 0;
}

ColumnData ResultSet::getARow(int row){
	return QueryData[row];
}

void ResultSet::addARow(ColumnData aRow){

	QueryData.push_back(aRow);
	rows++;
}

int ResultSet::setParameterType(PGresult *res)
{
	
	Oid type;
	int ret_val = 0;
	
	for(int i = 0; i < PQnfields(res) && ret_val == 0; ++i)
	{
		type = PQftype(res, i);
		
		switch(type){
		case 16 : // boolean
		   parameter_type[i] = META_LONG;
		   break;
		case 20 :
			parameter_type[i] = META_LONG;
			break;
		case 22 : 
			parameter_type[i] = META_LONG;
			break;
		case 21 : //int2
		   parameter_type[i] = META_LONG;
		   break;
		case 23 : //int4 
		   parameter_type[i] = META_LONG;
		   break;
		case 700 : //real 
		   parameter_type[i] = META_DOUBLE;
		   break;
		case 701 : // double precision 
		   parameter_type[i] = META_DOUBLE;
		   break;
		case 1043 : // varchar
		   parameter_type[i] = META_STRING;
		   break;
		case 17 : //bytea
		   parameter_type[i] = META_STRING;
		   break;
		case 25 : //text
		   parameter_type[i] = META_STRING;
		   break;
		default :
		   fprintf(stderr,"ERROR: unknown data type!\n");
		   fflush(stderr); ret_val = -1;
		   break;
		}
	}

   return ret_val;
}


int ResultSet::setParameterName(PGresult *res)
{
        int ret_val = 0;
        char* name;
		string str;
		
        for(int i = 0; i < PQnfields(res) && ret_val == 0; ++i)
        {
		name = PQfname(res,i); str = name;
		Utility::stringToLower((char *)str.c_str(),str.size());
		
		column_name[str] = i;
		
        }
   return ret_val;
}

ResultSet& ResultSet::operator=(ResultSet const &other){
	if (&other != this){
		clear();
		QueryData = other.QueryData;
		
		parameter_type = (int *)malloc(sizeof(int)*other.columns);	
	
		for (int i=0; i< other.columns; i++)
			parameter_type[i] = other.parameter_type[i];
			
		reserved = other.reserved;
		column_name = other.column_name;
		rows = other.rows;
		columns = other.columns;
	}
	
	return *this;
}

/********* END of ResultSet ************/
ColumnData::~ColumnData(){
	clear();
}

void ColumnData::clear()
{
	doublearg.clear(); 
	longarg.clear();
	stringarg.clear();
}

int ColumnData::setElement(PGresult *res,int row, int column, int type)
{

	if(type == META_DOUBLE)
	{
		if(PQgetisnull(res,row,column))
		{
			fprintf(stderr,"ERROR: non existind data for row %d\
				column %d\n",row, column);fflush(stderr);
			doublearg[column] = NULL;
			return 0;
		}
		
		numeric *numval;
		if ((numval = PGTYPESnumeric_new()) == NULL){
			fprintf(stderr,"ERROR: can not create a new numeric object!\n");
			fflush(stderr);return -1;
		}
		 
		numval = PGTYPESnumeric_from_asc(PQgetvalue(res,row,column), NULL);
		
		double val;
		PGTYPESnumeric_to_double(numval, &val);
		PGTYPESnumeric_free(numval);
		 
		if(doublearg.find(column) != doublearg.end())
		{
			fprintf(stderr,"ERROR: results set is full!\n");
			fflush(stderr);return -1;
		}
		doublearg[column] = val;		
		
	}
	else if(type == META_LONG)
	{
		if(PQgetisnull(res,row,column))
		{
			fprintf(stderr,"ERROR: non existing data for row %d\
				column %d\n",row,column);fflush(stderr);
			longarg[column] = NULL;
			return 0;
		}
		
		numeric *numval;
		if ((numval = PGTYPESnumeric_new()) == NULL){
			fprintf(stderr,"ERROR: can not create a new numeric object!\n");
			fflush(stderr);return -1;
		}
		 
		numval = PGTYPESnumeric_from_asc(PQgetvalue(res,row,column), NULL);
		
		long val;
		PGTYPESnumeric_to_long(numval, &val);
		PGTYPESnumeric_free(numval);
		
		if(longarg.find(column) != longarg.end())
		{
			fprintf(stderr,"ERROR: results set is full!\n");
			fflush(stderr);return -1;

		}
		longarg[column] = val;
		
	}
	else if(type == META_STRING)
	{
		if(PQgetisnull(res,row,column))
		{
			fprintf(stderr,"ERROR: non existind data for row %d\
				column %d\n",row, column);fflush(stderr);
			stringarg[column].clear() ;
			return 0;
		}
		string val(PQgetvalue(res,row,column));
		//string val(Field->asString().GetBuffer(200));
		if(stringarg.find(column) != stringarg.end())
		{
			fprintf(stderr,"ERROR: results set is full!\n");
			fflush(stderr);return -1;
		}
		stringarg[column] = val;
	}
	else
	{
       	   	fprintf(stderr,"ERROR: Unknown type %d for column %d\n",
			type,column);fflush(stderr);
		return -1;
	}
	return 0;
}

int ColumnData::getElement(int column,double *res,int type)
{
	if(type == META_STRING)
	{
		fprintf(stderr, "ERROR: Parameter %d is of type string\n",
			column);
		fflush(stderr); return -1;
	}
		
	if(type == META_LONG)
	{
		fprintf(stderr, "WARNING: Parameter %d is of type long\n",
			column);
		fflush(stderr);
		long resaux;
		int ierr = getElementStrongType(column, &resaux);
		CHKERR(ierr); *res = resaux;
	}
	int ierr = getElementStrongType(column, res);
	return ierr; 

}
//----------------------------------------------------------------------
int ColumnData::getElement(int column,long *res,int type)
{
        if(type == META_STRING)
        {
                fprintf(stderr, "ERROR: Parameter %d is of type string\n",
			column);
                fflush(stderr); return -1;
        }

        if(type == META_DOUBLE)
        {
                fprintf(stderr, "WARNING: Parameter %d is of type double\n",
			column);
                fflush(stderr);
                double resaux;
                int ierr = getElementStrongType(column, &resaux);
                CHKERR(ierr); *res = (long)resaux;
        }
        int ierr = getElementStrongType(column, res);
        return ierr; 

}

int ColumnData::getElement(int column,string *res)
{
       	if(stringarg.find(column)==stringarg.end())
	{
		fprintf(stderr,"ERROR: no such column!\n");
		fflush(stderr);return -1;
	}
	*res =  stringarg[column];
	return 0;
}
//----------------------------------------------------------------------

int ColumnData::getElementStrongType(int column,long *res)
{
        if(longarg.find(column)==longarg.end())
	{
		fprintf(stderr,"ERROR: no such column!\n");
		fflush(stderr);return -1;
	}
	*res =  longarg[column];
	return 0;
}	

int ColumnData::getElementStrongType(int column,double *res)
{
        if(doublearg.find(column)==doublearg.end())
	{
		fprintf(stderr,"ERROR: no such column!\n");
		fflush(stderr);return -1;
	}
	*res =  doublearg[column];
	return 0;
}

	/******  END of ColumnData ***************/
}
