/**
 * \file Metadata.hxx Defines abstract Parameter, ParameterSet, and their implementation, DefaultProperty and PropertySet
 * \author Li Li
 * \date 4/30/08
 */

#ifndef Metadata_H_
#define Metadata_H_

#define META_INT	132
#define META_LONG   	133
#define META_DOUBLE 	134
#define META_STRING 	135
#define META_BOOLEAN    136

#define LT		1301
#define LTEQ	1302	
#define EQ		1303	
#define NEQ		1304	
#define GT		1305
#define GTEQ	1306
#define BT		1307

#define LEFT		110
#define RIGHT		111

// postgresql header
extern "C" {
#include "libpq-fe.h" 
}
#include <string>
#include <vector>
#include <map>
#include "Utility.hxx"
using namespace std;

namespace cqosdb 
{ 

	/**
	  *	This class defines a parameter, e.g., a parameter in {P} in meta-partitioner
	  */
	class Parameter { 
		public:
		virtual ~Parameter(){}
		
		/**
			Return class name
		*/
		virtual string getClassName()=0;
		
		/**
			Set parameter name
		*/
		virtual	void setName(string name)=0;
				
		/**
			Set parameter value, integer type
		*/
		virtual	void setValue(int value)=0;
		
		/**
			Set parameter value, long integer type
		*/
		virtual	void setValue(long value)=0;
		
		/**
			Set parameter value, double type
		*/
		virtual	void setValue(double value)=0;
		
		/**
			Set parameter value, string type
		*/
		virtual	void setValue(string value)=0;
		
		/**
			Set parameter value, boolean type
		*/
		virtual	void setValue(bool value)=0;
		
		/**
			Set table where the parameter is stored in the database.
		*/
		virtual void setTBName(string tbName)=0;
		
		/**
			Set trial the parameter belongs to
		*/
		virtual void setTrialID(int trialID)=0;
		
		/**
			Grab parameter name
		*/
		virtual	string getName()=0;
		
		/**
			Grab parameter integer value
		*/
		virtual	void getValue(int *value)=0;
		
		/**
			Grab parameter long integer value
		*/
		virtual	void getValue(long *value)=0;
		
		/**
			Grab parameter double value
		*/
		virtual	void getValue(double *value)=0;
		
		/**
			Grab parameter string value
		*/
		virtual	void getValue(string *value)=0;
		
		/**
			Grab parameter boolean value
		*/
		virtual	void getValue(bool *value)=0;
		
		/**
			Grab database table where the parameter is stored
		*/
		virtual string getTBName()=0;
		
		/**
			Grab trial the parameter belongs to
		*/
		virtual int getTrialID()=0;
			
		/**
			Compare with another parameter of the same type.
			\param another another parameter
			\param tolerance comparison tolerance value
			\param relation comparison relation, which could be LT, LTEQ, GT, GTEQ, EQ, NEQ, BT
			\return true if (this-another) .relation. tolerance is true,
					false otherwise
		*/
		virtual bool compareWith(Parameter& another, double tolerance, int relation)=0;
		
		/**
			Compare with another parameter of the same type.
			\param another another parameter
			\param relation comparison relation, which could be LT, LTEQ, GT, GTEQ, EQ, NEQ
			\return true if this .relation. another is true,
					false otherwise
		*/
		virtual bool compareWith(Parameter& another, int relation)=0;
		
		/**
		Produce a SQL string text that specifies conditions related to the parameter, which can be used in 'where' clause in a SQL command 
		for retrieving parameter value from DB.
		for example, if parameter.getName() = 'myName', parameter.getTBName() = 'myTB', parameter.getTrialID() =0; parameter.getInterEvtNo()=1,
		then the string is like:
		myTB.parameter_name='myName' and myTB.trial=0 and myTB.intervel_event=1
		*/
		virtual	string getDBConditions()=0;
		
		/**
		Produce a SQL string text that specifies conditions related to the parameter, which can be used in 'where' clause in a SQL command
		for retrieving parameters satisfying the conditions.
		\param tbIndex 
		>= 0 index of the parameter table
		< 0 don't use index
		\param value, 
		value to compare with
		\param relation comparison relation, 
		either L, LTEQ, EQ, NEQ, GT, GTEQ, or BT
		
		For example, if parameter.getName() = 'myName', parameter.getTBName() = 'myTB', parameter.getTrialID() =0; parameter.getInterEvtNo()=1, parameter.getValue()=3,
		tbIndex =0, value = 0.3, relation = LT
		then the string is like:
		myTB0.parameter_name='myName' and myTB0.trial=0 and myTB0.interval_event=1 and myTB0.parameter_value < 0.3		
		*/
		virtual	string getDBConditions(int tbIndex, double value, int relation)=0;
		
		/**
		Produce a SQL string text that specifies conditions related to the parameter, which can be used in 'where' clause in a SQL command
		for retrieving parameters satisfying the conditions.
		\param tbIndex 
		>= 0 index of the parameter table
		< 0 don't use index
		\param aStr 
		string to compare with
		\param relation comparison relation, 
		either EQ, or NEQ
		
		For example, if parameter.getName() = 'myName', parameter.getTBName() = 'myTB', parameter.getTrialID() =0; parameter.getInterEvtNo()=1, parameter.getValue()='foo',
		tbIndex =0, aStr = 'fet', relation = eq
		then the string is like:
		myTB0.parameter_name='myName' and myTB0.trial=0 and myTB0.interval_event=1 and myTB0.parameter_value = 'fet'		
		*/
		virtual	string getDBConditions(int tbIndex, string aStr, int relation)=0;
		
		/**
		Produce a SQL string text that specifies conditions related to the parameter, which can be used in 'where' clause in a SQL command
		for retrieving parameters satisfying the conditions.
		\param tbIndex 
		>= 0 index of the parameter table
		< 0 don't use index
		\param boolVal 
		boolean value to compare with
		\param relation comparison relation, 
		either EQ, or NEQ
		
		For example, if parameter.getName() = 'myName', parameter.getTBName() = 'myTB', parameter.getTrialID() =0; parameter.getInterEvtNo()=1, parameter.getValue()='true',
		tbIndex =0, boolVal = true, relation = eq
		then the string is like:
		myTB0.parameter_name='myName' and myTB0.trial=0 and myTB0.interval_event=1 and myTB0.parameter_value = 'true'		
		*/
		virtual	string getDBConditions(int tbIndex, bool boolVal, int relation)=0;
		
	};

	class ParameterSet { // e.g., {P}, {G}, {O} in meta-partitioner
		public:
		virtual ~ParameterSet(){}
		
		/**
			Return class name
		*/
		virtual string getClassName()=0;
		
		/**
			Add a parameter to the set
		*/
		virtual void addAParameter(Parameter& aParam)=0;
		
		/**
			Get a parameter in the set
			\param name Parameter name
		*/
		virtual Parameter* getParameterAt(string name)=0;
		
		/**
			Get a parameter in the set
			\param index Parameter index in the set
		*/
		virtual Parameter* getParameterAt(int index)=0;
		
		/**
			Get count of parameters in the set
		*/
		virtual	int getNumOfParameters()=0;
		
		/**
			Get table where the set is store in database
		*/
		virtual string getTBName()=0;
		
		/**
			Compare with another parameter set of the same type.
			\param another another parameter set
			\param tMap comparison tolerance values
			\param rMap comparison relations, which could be LT, LTEQ, GT, GTEQ, EQ, NEQ, BT
			\return true if every pair of parameters comparison is true 
					false otherwise
		*/
		virtual bool compareWith(ParameterSet& another, map<string, double> tMap, map<string, int> rMap)=0;
		
		/**
			Compare with another parameter set of the same type.
			\param another another parameter set
			\param rMap comparison relations, which could be LT, LTEQ, GT, GTEQ, EQ, NEQ
			\return true if every pair of parameters comparison is true 
					false otherwise
		*/
		virtual bool compareWith(ParameterSet& another, map<string, int> rMap)=0;
		
		/**
		Produce a SQL string text that specifies conditions related to the parameter set, which can be used in 'where' clause in a SQL command.
		\param tbIndices indices  of tables storing the parameter set
		\param tMap comparison tolerance values
		\param rMap comparison relations, either L, LTEQ, EQ, NEQ, GT, GTEQ, or BT
		*/

		virtual	string getDBConditions(int* tbIndices, map<string, double> tMap, map<string, int> rMap)=0;	
		
		virtual ParameterSet* clone()=0;
	};

	/**
		An implementation of Parameter.
	*/
	class DefaultProperty:
                public virtual Parameter
        {
        public:
		
		/**
		Constructor: initialize with parameter name, type and integer parameter value, the type should be 'int', 'long', 'double', 'string' or 'bool' 
		*/
		DefaultProperty(string name, string type, int value);
		
		/**
		Constructor: initialize with parameter name, type and long parameter value, the type should be 'int', 'long', 'double', 'string' or 'bool'
		*/
		DefaultProperty(string name, string type, long value);
		
		/**
		Constructor: initialize with parameter name, type and double parameter value, the type should be 'int', 'long', 'double', 'string' or 'bool' 
		*/
		DefaultProperty(string name, string type, double value);
		
		/**
		Constructor: initialize with parameter name, type and string parameter value, the type should be 'int', 'long', 'double', 'string' or 'bool' 
		*/
		DefaultProperty(string name, string type, string value);
		
		/**
		Constructor: initialize with parameter name, type and boolean parameter value, the type should be 'int', 'long', 'double', 'string' or 'bool' 
		*/
		DefaultProperty(string name, string type, bool value);
		
		/**
		Constructor: initialize with parameter name and type, leaving the value empty and to be filled in later, the type should be 'int', 'long', 'double', 'string' or 'bool'
		*/
		DefaultProperty(string name, string type);
		DefaultProperty();
		~DefaultProperty(){};
		string getClassName();

		void setName(string name);
		void setValue(int value);
		void setValue(long value);
		void setValue(double value);
		void setValue(string value);
		void setValue(bool value);
		void setTBName(string tbName);
		void setTrialID(int trialID);
		
		/**
			Set interval event(intended to be a phase event) the parameter is associated with
		*/
		void setInterEvt(string interEvt);
		
		/**
			Set parameter's category name.
		*/
		void setCatName(string catName);
		
		/**
			Set parameter's type
		*/
		void setParamType(string type);

		string getName();
		void getValue(int *value);
		void getValue(long *value);
		void getValue(double *value);
		void getValue(string *value);
		void getValue(bool *value);
		string getTBName();
		int getTrialID();
		
		/**
			Grab intervel event number
		*/
		string getInterEvt();
		
		/**
			Grab parameter's category name
		*/
		string getCatName();
		
		/**
			Grab parameter's type
		*/
		string getParamType();
		
		bool compareWith(Parameter& another, double tolerance, int relation);// if this is within 'tolerance' of another
		bool compareWith(Parameter& another, int relation);

		string getDBConditions(); // for retrieving the parameter value from DB
		string getDBConditions(int tbIndex, double tolerance, int relation); // for retrieving parameter value satisfying the conditions
		string getDBConditions(int tbIndex, string aStr, int relation);
		string getDBConditions(int tbIndex, bool boolVal, int relation);
		//DefaultProperty &operator=(const DefaultProperty& other);

	private:
	
		/**
			Correspond to parameter_name in the DB
		*/
		string propertyName; 
		
		/**
			Correspond to parameter_type in the DB
		*/
		string propertyType;
		
		/**
			Correspond to category_name in the DB
		*/
		string catName;
		
		/**
			Correspond to  parameter_value in the DB
		*/
		double doubleValue;
		string stringValue;
		bool boolValue;
		
		/**
			Supposed to be 'metadata_parameters'
		*/
		string tableName;
		
		/**
			Correspond to trial and interval_event in the DB
		*/
		int trialID;
		string interEvt;
		bool hasAValue;
	};
	

	/**
		An implementation of ParameterSet. A PropertySet object contains a set of distinct DefaultPropertys. 
	*/
	class PropertySet: 
		public virtual ParameterSet
	{
		
	public:
	
		/**
			Constructor: initialize with a vector of Parameters
		*/
		PropertySet(vector<DefaultProperty> & pVec);
		PropertySet(){};
		PropertySet(const PropertySet& other);
		~PropertySet();
		string getClassName();
		
		int getNumOfParameters();	
		string getTBName();		
		void addAParameter(Parameter& aParam);	
		// note: We save a copy of aParam object in the class. Original aParam is not modified in this function.
					
		Parameter* getParameterAt(string name);
		Parameter* getParameterAt(int index);
			
		bool compareWith(ParameterSet& another, map<string, double> tMap, map<string, int> rMap);
		bool compareWith(ParameterSet& another, map<string, int> rMap);

		string getDBConditions(int* tbIndices, map<string, double> tMap, map<string, int> rMap);
			
		PropertySet &operator=(const PropertySet& other);
		ParameterSet* clone();
		
		/**
			A vector of parameters
		*/
		vector<DefaultProperty> paramVec;
	};

	/**
	Produce a SQL string text that describes cross join conditions of two sets of parameters. 
	For example, if set1's table name = app, set1={param1, param2, param3}, set2's table name = exp, set2={param1, param2, param3'}. 
	then the crossjoin string is like:
	app.param1=exp.param1 and app.param2=exp.param2
	*/
	string crossJoinStr(ParameterSet& set1, ParameterSet& set2);

	/**
	Trim a string and trun it to lower case;
	*/
	void normString(string& type);
					
}  // namespace cqosdb
#endif


